#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
import re
import select
import time
import traceback
from typing import List, Dict, Any

import paramiko

from fabric_am.util.am_constants import AmConstants
from fabric_am.util.ansible_helper import AnsibleHelper


class Utils:
    """
    Helper class to basic support functions
    """
    @staticmethod
    def parse_vcpuinfo(*, vcpuinfo_output: List[str]) -> List[Dict[str, str]]:
        """
        Parse VcpuInfo and transform it to JSON
        @param vcpuinfo_output: vcpuinfo command output
        """
        '''
        Vcpuinfo Output looks like:
            cpu_info = ["VCPU:           0",
                "CPU:            27",
                "State:          running",
                "CPU time:       3.1s",
                "CPU Affinity:   yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy", 
                "", 
                "VCPU:           1", 
                "CPU:            9", 
                "State:          running", 
                "CPU Affinity:   yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy"
                ]
        '''
        result = []
        data = {}
        for line in vcpuinfo_output:
            if line == "":
                result.append(data)
                data.clear()
            else:
                # Split the line at the first occurrence of ':' character
                key, value = line.split(':', 1)
                # Remove leading/trailing spaces from key and value
                key = key.strip()
                value = value.strip()
                # Add the key-value pair to the data dictionary
                data[key] = value
        return result

    @staticmethod
    def parse_lscpu(*, lscpu_output: List[str]) -> List[Dict[str, str]]:
        """
        Parse Lscpu and transform it to JSON
        @param lscpu_output: lscpu command output
        """
        '''
        ['CPU NODE', '0   0', '1   0', '2   0', '3   0', '4   0', '5   0', '6   0', '7   0', '8   1', '9   1', '10  1',
         '11  1', '12  1', '13  1', '14  1', '15  1', '16  2', '17  2', '18  2', '19  2', '20  2', '21  2', '22  2', 
         '23  2', '24  3', '25  3', '26  3', '27  3', '28  3', '29  3', '30  3', '31  3', '32  4', '33  4', '34  4', 
         '35  4', '36  4', '37  4', '38  4', '39  4', '40  5', '41  5', '42  5', '43  5', '44  5', '45  5', '46  5', 
         '47  5', '48  6', '49  6', '50  6', '51  6', '52  6', '53  6', '54  6', '55  6', '56  7', '57  7', '58  7', 
         '59  7', '60  7', '61  7', '62  7', '63  7', '64  0', '65  0', '66  0', '67  0', '68  0', '69  0', '70  0', 
         '71  0', '72  1', '73  1', '74  1', '75  1', '76  1', '77  1', '78  1', '79  1', '80  2', '81  2', '82  2', 
         '83  2', '84  2', '85  2', '86  2', '87  2', '88  3', '89  3', '90  3', '91  3', '92  3', '93  3', '94  3', 
         '95  3', '96  4', '97  4', '98  4', '99  4', '100 4', '101 4', '102 4', '103 4', '104 5', '105 5', '106 5', 
         '107 5', '108 5', '109 5', '110 5', '111 5', '112 6', '113 6', '114 6', '115 6', '116 6', '117 6', '118 6', 
         '119 6', '120 7', '121 7', '122 7', '123 7', '124 7', '125 7', '126 7', '127 7']
        '''
        result = []
        for line in lscpu_output:
            if line == "" or "CPU" in line:
                continue
            else:
                data = {}
                # Split the line at the first occurrence of ' ' character
                cpu, node = line.split(' ', 1)
                # Remove leading/trailing spaces from key and value
                cpu = cpu.strip()
                node = node.strip()
                # Add the key-value pair to the data dictionary
                data['cpu'] = cpu
                data['node'] = node
                result.append(data)
        return result

    @staticmethod
    def parse_numastat(*, numastat_output: str) -> Dict[Any, dict]:
        """
        Parse numa stat output and convert it to JSON
        @param numastat_output: Numa State output
        """
        '''
        Numa Stat Output looks like:
        numa_info = ["",
                 "Per-node process memory usage (in MBs) for PID 2676600 (qemu-kvm)",
                 "         Node 0 Node 1 Total",
                 "         ------ ------ -----",
                 "Private   17531  15284 32815",
                 "Heap          0      6     6",
                 "Stack         0      0     0",
                 "Huge          0      0     0",
                 "-------  ------ ------ -----",
                 "Total     17531  15290 32821"
                 ]
        '''
        result = {}
        keys = []
        for line in numastat_output:
            # Ignore empty lines and headers
            if line == "" or "Per-node" in line or "--" in line:
                continue
            # Extract Numa Nodes
            elif "Node" in line:
                line = line.strip()
                regex_pattern = r'\b(\w+\s+\d+)\b|\b(\w+)\b'
                matches = re.findall(regex_pattern, line)
                # Flatten the list of tuples and remove empty strings
                keys = [match[0] or match[1] for match in matches if match[0] or match[1]]

                # Add empty dictionary for each Node and Total
                for x in keys:
                    result[x] = {}
            else:
                # Extract Different types of Memory for each Node and Total
                regex_pattern = r'(\b\w+\b)|(\b\d+\b)'
                matches = re.findall(regex_pattern, line)
                # Flatten the list of tuples and remove empty strings
                memory_values = [match[0] or match[1] for match in matches if match[0] or match[1]]
                # Index 0 contains the type of Memory and is used as a key
                idx = 1
                for x in keys:
                    result[x][memory_values[0]] = memory_values[idx]
                    idx += 1

        return result

    @staticmethod
    def parse_numactl(*, numactl_output: str) -> Dict[Any, dict]:
        """
        Parse numa stat output and convert it to JSON
        @param numastat_output: Numa State output
        """
        '''
        Numa Stat Output looks like:
        numa_info = [
                     'available: 8 nodes (0-7)', 
                     'node 0 cpus: 0 1 2 3 4 5 6 7 64 65 66 67 68 69 70 71', 
                     'node 0 size: 63798 MB', 
                     'node 0 free: 59660 MB', 
                     'node 1 cpus: 8 9 10 11 12 13 14 15 72 73 74 75 76 77 78 79', 
                     'node 1 size: 64507 MB', 
                     'node 1 free: 63042 MB', 
                     'node 2 cpus: 16 17 18 19 20 21 22 23 80 81 82 83 84 85 86 87', 
                     'node 2 size: 64507 MB', 
                     'node 2 free: 63911 MB', 
                     'node 3 cpus: 24 25 26 27 28 29 30 31 88 89 90 91 92 93 94 95', 
                     'node 3 size: 64458 MB', 
                     'node 3 free: 30882 MB', 
                     'node 4 cpus: 32 33 34 35 36 37 38 39 96 97 98 99 100 101 102 103', 
                     'node 4 size: 64507 MB', 
                     'node 4 free: 64069 MB', 
                     'node 5 cpus: 40 41 42 43 44 45 46 47 104 105 106 107 108 109 110 111', 
                     'node 5 size: 64507 MB', 
                     'node 5 free: 63966 MB', 
                     'node 6 cpus: 48 49 50 51 52 53 54 55 112 113 114 115 116 117 118 119', 
                     'node 6 size: 64507 MB', 
                     'node 6 free: 61214 MB', 
                     'node 7 cpus: 56 57 58 59 60 61 62 63 120 121 122 123 124 125 126 127', 
                     'node 7 size: 64506 MB', 
                     'node 7 free: 63869 MB', 
                     'node distances:', 
                     'node   0   1   2   3   4   5   6   7 ', 
                     '  0:  10  12  12  12  32  32  32  32 ', 
                     '  1:  12  10  12  12  32  32  32  32 ', 
                     '  2:  12  12  10  12  32  32  32  32 ', 
                     '  3:  12  12  12  10  32  32  32  32 ', 
                     '  4:  32  32  32  32  10  12  12  12 ', 
                     '  5:  32  32  32  32  12  10  12  12 ', 
                     '  6:  32  32  32  32  12  12  10  12 ', 
                     '  7:  32  32  32  32  12  12  12  10 '
                     ]
        '''
        result = {}
        data = {}
        for line in numactl_output:
            # Ignore empty lines and headers
            if line == "":
                continue
            elif "available" in line:
                key, value = line.split(':')
                key = key.strip()
                value = value.strip()
                result[key] = value
            elif "node distances" in line:
                break
            else:
                key, value = line.split(':')
                key = key.strip()
                value = value.strip()
                regex_pattern = r"^(.*?)\s+(\w+)$"
                matches = re.match(regex_pattern, key)
                if matches:
                    node = matches.group(1)
                    key = matches.group(2)
                    if node not in result:
                        result[node] = {}
                    result[node][key] = value
                else:
                    result[key] = value
        return result

    @staticmethod
    def execute_command(*, mgmt_ip: str, user: str, command: str, logger: logging.Logger,
                        timeout: int = 60, retry: int = 3, pwd: str = None, ssh_key_file: str = None):
        """
        Execute a command on the VM
        :param mgmt_ip Management IP to access the VM
        :param user Default Linux user to use for SSH/Ansible
        :param command Command to execute
        :param logger logger
        :param timeout Timeout in seconds
        :param retry Number of retries
        :param pwd password
        :param ssh_key_file ssh_key file
        :return:
        """
        for i in range(retry):
            ssh = None
            try:
                ssh = paramiko.SSHClient()
                ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

                if pwd is not None:
                    ssh.connect(mgmt_ip, username=user, password=pwd, timeout=timeout)
                else:
                    pkey = paramiko.RSAKey.from_private_key_file(ssh_key_file)
                    ssh.connect(mgmt_ip, username=user, timeout=timeout, pkey=pkey)

                # Execute command with timeout
                stdin, stdout, stderr = ssh.exec_command(command, timeout=timeout)

                # Ensure we don't hang forever
                while not stdout.channel.exit_status_ready():
                    rl, wl, xl = select.select([stdout.channel], [], [], timeout)
                    if not rl:
                        raise TimeoutError(f"Command timeout: {command}")

                output = stdout.read().decode().strip()
                error = stderr.read().decode().strip()

                ssh.close()

                if error:
                    logger.error(f"Error executing command: {error}")
                    return None
                return output

            except TimeoutError as e:
                logger.error(f"Command timeout: {command}")
                ssh.close()
                raise e
            except Exception as e:
                logger.error(f"Exception: {e}")
                logger.error(traceback.format_exc())
                if i < retry - 1:
                    time.sleep(timeout)
                    logger.info(f"Retrying command {command} on VM {mgmt_ip}")
                else:
                    logger.error(f"Failed to execute command {command} on VM {mgmt_ip}")
                    raise e
            finally:
                if ssh:
                    ssh.close()

    @staticmethod
    def verify_ssh(*, mgmt_ip: str, user: str, logger: logging.Logger, timeout: int = 60, retry: int = 10,
                   pwd: str = None, ssh_key_file: str = None):
        """
        Verify that the VM is accessible via SSH
        :param mgmt_ip Management IP to access the VM
        :param user Default Linux user to use for SSH/Ansible
        :param logger Logger
        :param timeout timeout
        :param retry Number of retries
        :param pwd password
        :param ssh_key_file ssh_key_file

        """
        command = "echo 'SSH connection verified'"

        try:
            output = Utils.execute_command(mgmt_ip=mgmt_ip, user=user, command=command, logger=logger,
                                           timeout=timeout, retry=retry, pwd=pwd, ssh_key_file=ssh_key_file)
            if output:
                logger.info(f"SSH Verification Output: {output}")
                return True
        except Exception:
            return False

    @staticmethod
    def execute_ansible(*, inventory_path: str, playbook_path: str, extra_vars: dict, logger: logging.Logger,
                        sources: str = None, private_key_file: str = None, host_vars: dict = None,
                        host: str = None, user: str = None):
        """
        Execute ansible
        :param inventory_path: inventory location
        :param playbook_path: playbook
        :param extra_vars: extra vars
        :param logger: logger
        :param sources: sources
        :param private_key_file: private key file
        :param host_vars: host vars
        :param host: host
        :param user: user

        :return OK results
        :raises Exception in case of failure
        """
        ansible_helper = AnsibleHelper(inventory_path=inventory_path, logger=logger, sources=sources)

        ansible_helper.set_extra_vars(extra_vars=extra_vars)

        if host is not None and host_vars is not None and len(host_vars) > 0:
            for key, value in host_vars.items():
                ansible_helper.add_vars(host=host, var_name=key, value=value)

        logger.info(f"Executing playbook {playbook_path} extra_vars: {extra_vars} host_vars: {host_vars}")
        ansible_helper.run_playbook(playbook_path=playbook_path, private_key_file=private_key_file, user=user)
        return ansible_helper.get_result_callback().get_json_result_ok()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org), Xi Yang (xiwang@es.net), Komal Thareja (kthare10@renci.org)
import logging
import threading
import unittest
import time
import uuid

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.core.unit import Unit
from fabric_cf.actor.core.util.id import ID
from fim.slivers.attached_components import ComponentSliver, ComponentType, AttachedComponentsInfo
from fim.slivers.capacities_labels import Capacities, Labels, CapacityHints
from fim.slivers.instance_catalog import InstanceCatalog
from fim.slivers.network_service import NetworkServiceSliver, ServiceType, NSLayer, MirrorDirection
from fim.slivers.interface_info import InterfaceSliver, InterfaceType, InterfaceInfo
from fim.slivers.path_info import Path, PathInfo, ERO
from fim.slivers.gateway import Gateway

# FIXME: @xiyang
from fabric_am.handlers.al2s_handler import Al2sHandler
from fabric_am.util.am_constants import AmConstants

class TestAl2sHandler(unittest.TestCase):
    logger = logging.getLogger(__name__)
    log_format = \
        '%(asctime)s - %(name)s - {%(filename)s:%(lineno)d} - [%(threadName)s] - %(levelname)s - %(message)s'
    logging.basicConfig(handlers=[logging.StreamHandler()], format=log_format, force=True)

    def setUp(self) -> None:
        self.unit = Unit(rid=ID(uid="rid-1"))
    

    def test_L3Cloud_GCP(self):
        # create a NetworkService sliver for FABNetv4
        prop = {AmConstants.CONFIG_PROPERTIES_FILE: '../config/al2s_handler_config.yml'}
    
        handler = Al2sHandler(logger=self.logger, properties=prop, process_lock=threading.Lock())
        #
        # create a network sliver for FABNetv4 and its interfaces
        #
        sliver = NetworkServiceSliver()
        # service name (set by user) - only guaranteed unique within a slice
        sliver.set_name('L3VPN-Cloud')
        # if service name global uniqueness is a requirement use Labels.local_name for that (optional)
        # e.g. concatenate name + res id (or another unique id)
        # sliver.set_labels(Labels().set_fields(local_name='test-l2bridge-shortname'))
        # per @xiyang he uses unit id for service name so this is not needed.
        sliver.set_type(ServiceType.L3VPN)
        sliver.set_layer(NSLayer.L3)
        # the ASN of *this* service
        sliver.set_labels(Labels(asn='55038', local_name='al2s_l3_gcp_interconn_test'))
    
        # this is the gateway with the IP range picked for this sliver in this slice on this site
        # can also be specified with ipv6/ipv6_subnet and mac is optional for both.
        # Q: does that mean that the advertisement needs to maintain information about multiple
        # subnet, gateway and mac tuples for each site?
        # sliver.set_gateway(Gateway(Labels(ipv4="10.128.128.254", ipv4_subnet="10.128.128.0/24")))
    
        #
        # create a small number of Interface slivers, set their properties and link to service
        #
    
        #
        # First interface
        #
        isl1 = InterfaceSliver()
        # the name is normally set by FIM as '-' concatenation of service name
        isl1.set_name('Google Cloud Platform')
        # this will be a ServicePort in the network service sliver. It is created by FIM automatically when
        # the user adds a NetworkService to the ASM. The name is set by the FIM as '-' concatenation of service
        # name and peer interface sliver name.
        isl1.set_type(InterfaceType.ServicePort)
    
        #
        # NOTE: use prefix with suffient size to assure local and peer on the same subnet
        #
        sliver_labels = Labels(ipv4_subnet='192.168.100.10/16',
                               vlan='100', 
                               local_name='Bundle-Ether5',
                               device_name='agg3.ashb')
        sliver_peer_labels = Labels(ipv4_subnet='192.168.100.20/16', asn='16550',
                                    # bgp_key='0xzsEwC7xk6c1fK_h.xHyAdx',
                                    account_id='18af373b-9f08-4cd1-a45a-5e2f570ece92/us-east4/2',
                                    local_name='Google Cloud Platform')
    
        # capacities (bw in Gbps, burst size is in Mbytes) source: (b)
        sliver_capacities = Capacities(bw=50, mtu=1460)
    
        # assign labels and capacities
        isl1.set_labels(sliver_labels)
        isl1.set_peer_labels(sliver_peer_labels)
        isl1.set_capacities(sliver_capacities)
    
        #
        # Second interface (let's assume this is a dedicated card)
        #
        isl2 = InterfaceSliver()
        isl2.set_name('FABRIC')
        isl2.set_type(InterfaceType.ServicePort)
    
        sliver_labels = Labels(ipv4_subnet='192.168.100.2/30',
                               vlan='100', 
                               local_name='HundredGigE0/0/0/27',
                               device_name='core1.star')
        sliver_peer_labels = Labels(ipv4_subnet='192.168.100.1/30', asn='64512',
                                    bgp_key='0xzsEwC7xk6c1fK_h.xHyAdx',
                                    local_name='FABRIC')
    
        # capacities (bw in Gbps, burst size is in Mbytes) source: (b)
        sliver_capacities = Capacities(bw=0, mtu=9000)
    
        isl2.set_labels(sliver_labels)
        isl2.set_peer_labels(sliver_peer_labels)
        isl2.set_capacities(sliver_capacities)
    
        # create interface info object, add populated interfaces to it
        ifi = InterfaceInfo()
        ifi.add_interface(isl1)
        ifi.add_interface(isl2)
    
        # add interface info object to sliver. All of this happens automagically normally
        sliver.interface_info = ifi
    
        # set a fake unit reservation
        uid = uuid.uuid3(uuid.NAMESPACE_DNS, 'test_L3Cloud')
        self.unit = Unit(rid=ID(uid=str(uid)))
        self.unit.set_sliver(sliver=sliver)
    
        #
        # create a service (create needs to parse out sliver information
        # into exact parameters the service ansible script needs)
        #
        r, updated_unit = handler.create(unit=self.unit)
        self.assertEqual(r[Constants.PROPERTY_TARGET_NAME], Constants.TARGET_CREATE)
        self.assertEqual(r[Constants.PROPERTY_ACTION_SEQUENCE_NUMBER], 0)
        self.assertEqual(r[Constants.PROPERTY_TARGET_RESULT_CODE], Constants.RESULT_CODE_OK)
    
        time.sleep(60)
    
        #
        # delete - need to make sure the updated unit has the right info to delete the service
        #
        r, updated_unit = handler.delete(updated_unit)
        self.assertEqual(r[Constants.PROPERTY_TARGET_NAME], Constants.TARGET_DELETE)
        self.assertEqual(r[Constants.PROPERTY_ACTION_SEQUENCE_NUMBER], 0)
        self.assertEqual(r[Constants.PROPERTY_TARGET_RESULT_CODE], Constants.RESULT_CODE_OK)

    
    def test_L3Cloud_AWS(self):
        # create a NetworkService sliver for FABNetv4
        prop = {AmConstants.CONFIG_PROPERTIES_FILE: '../config/al2s_handler_config.yml'}
    
        handler = Al2sHandler(logger=self.logger, properties=prop, process_lock=threading.Lock())
        #
        # create a network sliver for FABNetv4 and its interfaces
        #
        sliver = NetworkServiceSliver()
        # service name (set by user) - only guaranteed unique within a slice
        sliver.set_name('L3VPN-Cloud')
        # if service name global uniqueness is a requirement use Labels.local_name for that (optional)
        # e.g. concatenate name + res id (or another unique id)
        # sliver.set_labels(Labels().set_fields(local_name='test-l2bridge-shortname'))
        # per @xiyang he uses unit id for service name so this is not needed.
        sliver.set_type(ServiceType.L3VPN)
        sliver.set_layer(NSLayer.L3)
        # the ASN of *this* service
        sliver.set_labels(Labels(asn='55038', local_name='al2s_l3_aws_interconn_test'))
    
        # this is the gateway with the IP range picked for this sliver in this slice on this site
        # can also be specified with ipv6/ipv6_subnet and mac is optional for both.
        # Q: does that mean that the advertisement needs to maintain information about multiple
        # subnet, gateway and mac tuples for each site?
        # sliver.set_gateway(Gateway(Labels(ipv4="10.128.128.254", ipv4_subnet="10.128.128.0/24")))
    
        #
        # create a small number of Interface slivers, set their properties and link to service
        #
    
        #
        # First interface
        #
        isl1 = InterfaceSliver()
        # the name is normally set by FIM as '-' concatenation of service name
        isl1.set_name('AWS')
        # this will be a ServicePort in the network service sliver. It is created by FIM automatically when
        # the user adds a NetworkService to the ASM. The name is set by the FIM as '-' concatenation of service
        # name and peer interface sliver name.
        isl1.set_type(InterfaceType.ServicePort)
    
        sliver_labels = Labels(ipv4_subnet='192.168.5.2/24',
                               vlan='31', 
                               local_name='TenGigE0/0/0/13/3',
                               device_name='agg3.dall3')
        sliver_peer_labels = Labels(ipv4_subnet='192.168.5.1/24', asn='64512',
                                    bgp_key='0xzsEwC7xk6c1fK_h.xHyAdx', account_id='296256999979',
                                    local_name='AWS')
    
        # capacities (bw in Mbps, burst size is in Mbytes) source: (b)
        sliver_capacities = Capacities(bw=50, mtu=9001)
    
        # assign labels and capacities
        isl1.set_labels(sliver_labels)
        isl1.set_peer_labels(sliver_peer_labels)
        isl1.set_capacities(sliver_capacities)
    
        #
        # Second interface (let's assume this is a dedicated card)
        #
        isl2 = InterfaceSliver()
        isl2.set_name('FABRIC')
        isl2.set_type(InterfaceType.ServicePort)
    
        sliver_labels = Labels(ipv4_subnet='192.168.100.2/24',
                               vlan='854', 
                               local_name='HundredGigE0/0/0/24',
                               device_name='core1.loui')
        sliver_peer_labels = Labels(ipv4_subnet='192.168.100.1/24', asn='64512',
                                    bgp_key='0xzsEwC7xk6c1fK_h.xHyAdx',
                                    local_name='FABRIC')
    
        # capacities (bw in Gbps, burst size is in Mbytes) source: (b)
        sliver_capacities = Capacities(bw=0, mtu=9000)
    
        isl2.set_labels(sliver_labels)
        isl2.set_peer_labels(sliver_peer_labels)
        isl2.set_capacities(sliver_capacities)
    
        # create interface info object, add populated interfaces to it
        ifi = InterfaceInfo()
        ifi.add_interface(isl1)
        ifi.add_interface(isl2)
    
        # add interface info object to sliver. All of this happens automagically normally
        sliver.interface_info = ifi
    
        # set a fake unit reservation
        uid = uuid.uuid3(uuid.NAMESPACE_DNS, 'test_L3Cloud')
        self.unit = Unit(rid=ID(uid=str(uid)))
        self.unit.set_sliver(sliver=sliver)
    
        #
        # create a service (create needs to parse out sliver information
        # into exact parameters the service ansible script needs)
        #
        r, updated_unit = handler.create(unit=self.unit)
        self.assertEqual(r[Constants.PROPERTY_TARGET_NAME], Constants.TARGET_CREATE)
        self.assertEqual(r[Constants.PROPERTY_ACTION_SEQUENCE_NUMBER], 0)
        self.assertEqual(r[Constants.PROPERTY_TARGET_RESULT_CODE], Constants.RESULT_CODE_OK)
        
        time.sleep(30)
        
        #
        # modify -
        #

        modified_sliver = NetworkServiceSliver()
        # service name (set by user) - only guaranteed unique within a slice
        modified_sliver.set_name('L3VPN-Cloud')  # must be the same name
        modified_sliver.set_type(ServiceType.L3VPN)
        modified_sliver.set_layer(NSLayer.L3)

        ifi = InterfaceInfo()  # started over
        
        # remove interface 1
        # ifi.add_interface(isl1)  # Add Interface1 (reused)

        # modify Interface 2
        isl2 = InterfaceSliver()
        isl2.set_name('FABRIC')
        isl2.set_type(InterfaceType.ServicePort)
        
        sliver_labels = Labels(ipv4_subnet='192.168.100.20/24',
                               vlan='854', 
                               local_name='HundredGigE0/0/0/24',
                               device_name='core1.loui')
        sliver_peer_labels = Labels(ipv4_subnet='192.168.100.10/24', asn='64512',
                                    bgp_key='0xzsEwC7xk6c1fK_h.xHyAdx',
                                    local_name='FABRIC')
        
        # capacities (bw in Gbps, burst size is in Mbytes) source: (b)
        sliver_capacities = Capacities(bw=0, mtu=9000)
        
        isl2.set_labels(sliver_labels)
        isl2.set_peer_labels(sliver_peer_labels)
        isl2.set_capacities(sliver_capacities)
        
        ifi.add_interface(isl2)
        
        
        
        # add Interface 3
        isl3 = InterfaceSliver()
        isl3.set_name('FABRIC1')
        isl3.set_type(InterfaceType.ServicePort)
        
        sliver_labels = Labels(ipv4_subnet='192.168.100.25/24',
                               vlan='855', 
                               local_name='HundredGigE0/0/0/24',
                               device_name='core1.loui')
        sliver_peer_labels = Labels(ipv4_subnet='192.168.100.15/24', asn='64512',
                                    bgp_key='0xzsEwC7xk6c1fK_h.xHyAdx',
                                    local_name='FABRIC')
        
        # capacities (bw in Gbps, burst size is in Mbytes) source: (b)
        sliver_capacities = Capacities(bw=0, mtu=9000)
        
        isl3.set_labels(sliver_labels)
        isl3.set_peer_labels(sliver_peer_labels)
        isl3.set_capacities(sliver_capacities)
        
        ifi.add_interface(isl3)
        
        modified_sliver.interface_info = ifi

        updated_unit.set_modified(modified=modified_sliver)
        
        r, updated_unit = handler.modify(updated_unit)
        self.assertEqual(r[Constants.PROPERTY_TARGET_NAME], Constants.TARGET_MODIFY)
        self.assertEqual(r[Constants.PROPERTY_ACTION_SEQUENCE_NUMBER], 0)
        self.assertEqual(r[Constants.PROPERTY_TARGET_RESULT_CODE], Constants.RESULT_CODE_OK)

        time.sleep(30)
        
    
        #
        # delete - need to make sure the updated unit has the right info to delete the service
        #
        r, updated_unit = handler.delete(updated_unit)
        self.assertEqual(r[Constants.PROPERTY_TARGET_NAME], Constants.TARGET_DELETE)
        self.assertEqual(r[Constants.PROPERTY_ACTION_SEQUENCE_NUMBER], 0)
        self.assertEqual(r[Constants.PROPERTY_TARGET_RESULT_CODE], Constants.RESULT_CODE_OK)
    




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
import multiprocessing

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.core.unit import Unit
from fabric_cf.actor.core.util.id import ID
from fim.slivers import InstanceCatalog
from fim.slivers.attached_components import ComponentSliver, ComponentType, AttachedComponentsInfo
from fim.slivers.capacities_labels import Capacities, Labels, CapacityHints
from fim.slivers.network_node import NodeSliver, NodeType

from fabric_am.handlers.switch_handler import SwitchHandler
from fabric_am.handlers.vm_handler import VMHandler
from fabric_am.util.am_constants import AmConstants
from fabric_am.util.ansible_helper import AnsibleHelper


class TestPlaybooks:
    logger = logging.getLogger(__name__)
    log_format = \
        '%(asctime)s - %(name)s - {%(filename)s:%(lineno)d} - [%(threadName)s] - %(levelname)s - %(message)s'
    logging.basicConfig(handlers=[logging.StreamHandler()], format=log_format, force=True)
    logger.setLevel("DEBUG")

    prop = {AmConstants.CONFIG_PROPERTIES_FILE: 'fabric_am/config/vm_handler_config.yml'}
    lock = multiprocessing.Lock()
    handler = VMHandler(logger=logger, properties=prop, process_lock=lock)
    sw_handler = SwitchHandler(logger=logger, process_lock=lock,
                               properties={AmConstants.CONFIG_PROPERTIES_FILE: 'fabric_am/config/switch_handler_config.yml'})
    from fabric_cf.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().log = logger

    @staticmethod
    def create_unit(include_pci: bool = True, include_image: bool = True, include_name: bool = True,
                    include_instance_name: bool = False) -> Unit:
        """
        Create a unit
        :param include_pci:
        :param include_image:
        :param include_name:
        :param include_instance_name:
        :return:
        """
        u = Unit(rid=ID(uid='0a0c2fb9-071a-4a3a-ba94-aa178c237aa2'))
        sliver = NodeSliver()
        cap = Capacities(core=2, ram=8, disk=10)
        sliver.set_properties(type=NodeType.VM, site="UKY", capacity_allocations=cap)
        sliver.label_allocations = Labels(instance_parent="uky-w1.fabric-testbed.net")
        catalog = InstanceCatalog()
        instance_type = catalog.map_capacities_to_instance(cap=cap)
        cap_hints = CapacityHints(instance_type=instance_type)
        sliver.set_properties(capacity_hints=cap_hints,
                              capacity_allocations=catalog.get_instance_capacities(instance_type=instance_type))

        if include_name:
            sliver.set_properties(name="uky-w1-b")

        if include_image:
            sliver.set_properties(image_type='qcow2', image_ref='default_fedora')

        if include_pci:
            component = ComponentSliver()
            labels = Labels(bdf=["0000:41:00.0", "0000:41:00.1"])
            component.set_properties(type=ComponentType.SmartNIC, model='ConnectX-5', name='nic1',
                                     label_allocations=labels)
            sliver.attached_components_info = AttachedComponentsInfo()
            sliver.attached_components_info.add_device(device_info=component)

        if include_instance_name:
            sliver.label_allocations.instance = "instance-00001077"

        u.set_sliver(sliver=sliver)
        return u

    def test_create_vm_success(self):
        """
        Test successful creation of VM with PCI devices
        :return:
        """
        u = self.create_unit()

        r, u = self.handler.create(unit=u)
        print(r)
        print(u.sliver)
        return u

    def test_create_vm_success_no_pci(self):
        """
        Test successful creation of VM without PCI devices
        :return:
        """
        u = self.create_unit(include_pci=False)

        r, u = self.handler.create(unit=u)
        print(r)
        print(u.sliver)

    def test_delete_vm_success(self, u):
        """
        Test successful deletion of a VM
        :return:
        """
        if u is None:
            u = self.create_unit(include_instance_name=True, include_name=True)

        r, u = self.handler.delete(unit=u)
        print(r)
        print(u.get_sliver())

    def test_delete_vm_success_no_pci(self):
        """
        Test successful deletion of a VM without PCI devices
        :return:
        """
        u = self.create_unit(include_pci=False)

        r, u = self.handler.delete(unit=u)
        print(r)
        print(u.get_sliver())

    def test_config_nw_interface(self):
        self.handler.configure_network_interface(mgmt_ip="128.163.179.50", user=AmConstants.CENTOS,
                                                 resource_type=str(ComponentType.SmartNIC),
                                                 mac_address="0C:42:A1:78:F8:04",
                                                 ipv4_address="192.168.11.2")

    def test_config_nw_interface_tagged(self):
        self.handler.configure_network_interface(mgmt_ip="128.163.179.50", user=AmConstants.CENTOS,
                                                 resource_type=str(ComponentType.SmartNIC),
                                                 mac_address="0C:42:A1:78:F8:04",
                                                 ipv4_address="192.168.11.2", vlan="200")

    def test_poa_cpuinfo(self):
        u = self.create_unit(include_instance_name=True, include_name=True)
        data = {"operation": "cpuinfo"}
        self.handler.poa(unit=u, data=data)

    def test_poa_numainfo(self):
        u = self.create_unit(include_instance_name=True, include_name=True)
        data = {"operation": "numainfo"}
        self.handler.poa(unit=u, data=data)

    def test_poa_numatune(self):
        u = self.create_unit(include_instance_name=True, include_name=True)
        data = {"node_set": ["1", "2", "3"], "operation": "numatune"}
        self.handler.poa(unit=u,data=data)

    def test_poa_reboot(self):
        u = self.create_unit(include_instance_name=True, include_name=True)
        data = {"operation": "reboot"}
        self.handler.poa(unit=u, data=data)

    def test_poa_cpupin(self):
        u = self.create_unit(include_instance_name=True, include_name=True)
        data = {"vcpu_cpu_map": [{"vcpu": 0, "cpu": 34}, {"vcpu": 1, "cpu": 35}], "operation": "cpupin"}
        self.handler.poa(unit=u, data=data)

    def test_fpga_prov(self):
        u = Unit(rid=ID(uid='0a0c2fb9-071a-4a3a-ba94-aa178c237aa2'))
        u.properties = {Constants.USER_SSH_KEY:
                        "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAkoLn9RFJIJEHPhocIIcPl komalthareja@dhcp152-54-6-178.wireless.europa.renci.org,"
                        "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAPNnEFquLWMmkoLEHPhocIIcPl komalthareja@dhcp152-54-6-178.wireless.europa.renci.org"}
        sliver = NodeSliver()
        cap = Capacities(core=2, ram=8, disk=10)
        sliver.set_properties(type=NodeType.VM, site="RENC", capacity_allocations=cap, name="fpga-vm")
        sliver.label_allocations = Labels(instance_parent="renc-w2.fabric-testbed.net")
        catalog = InstanceCatalog()
        instance_type = catalog.map_capacities_to_instance(cap=cap)
        cap_hints = CapacityHints(instance_type=instance_type)
        sliver.set_properties(capacity_hints=cap_hints,
                              capacity_allocations=catalog.get_instance_capacities(instance_type=instance_type))
        sliver.set_properties(image_type='qcow2', image_ref='docker_rocky_8')

        component = ComponentSliver()
        labels = Labels(bdf=["0000:25:00.0", "0000:25:00.1"])
        component.set_properties(type=ComponentType.FPGA, model='Xilinx', name='nic1',
                                 label_allocations=labels, labels=labels)
        sliver.attached_components_info = AttachedComponentsInfo()
        sliver.attached_components_info.add_device(device_info=component)
        u.set_sliver(sliver=sliver)
        r, u = self.handler.create(unit=u)

    def test_switch(self):
        u = Unit(rid=ID(uid='0a0c2fb9-071a-4a3a-ba94-aa178c237aa2'))
        u.properties = {Constants.USER_SSH_KEY:
                            "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDIxGyj1Bma9BJEHPhocIIcPl komalthareja@dhcp152-54-6-178.wireless.europa.renci.org"}
        sliver = NodeSliver()
        cap = Capacities(unit=1)
        sliver.set_properties(type=NodeType.Switch, site="RENC", capacity_allocations=cap, name="p4-sw")
        u.sliver = sliver
        r, u = self.sw_handler.create(unit=u)


def execute_ansible(*, logger, inventory_path: str, playbook_path: str, extra_vars: dict,
                    ansible_python_interpreter: str, sources: str = None, private_key_file: str = None,
                    host_vars: dict = None, host: str = None, user: str = None):
    ansible_helper = AnsibleHelper(inventory_path=inventory_path, logger=logger,
                                   ansible_python_interpreter=ansible_python_interpreter,
                                   sources=sources)

    ansible_helper.set_extra_vars(extra_vars=extra_vars)

    if host is not None and host_vars is not None and len(host_vars) > 0:
        for key, value in host_vars.items():
            ansible_helper.add_vars(host=host, var_name=key, value=value)

    logger.info(f"Executing playbook {playbook_path} extra_vars: {extra_vars} host_vars: {host_vars}")
    ansible_helper.run_playbook(playbook_path=playbook_path, private_key_file=private_key_file, user=user)
    return ansible_helper.get_result_callback()


def test_audit():
    logger = logging.getLogger("Audit")
    results_1 = execute_ansible(inventory_path="/etc/fabric/actor/playbooks//inventory",
                                logger=logger,
                                playbook_path="/etc/fabric/actor/playbooks//worker_libvirt_operations.yml",
                                extra_vars={"operation": "listall"},
                                ansible_python_interpreter='/usr/bin/python3.6')

    # Dictionary to store OpenStack VMs
    os_vms = {}

    for host, ok_result in results_1.host_ok.items():
        # Get VMs via Virsh
        print(f"host: {host}")
        virsh_vms = []
        if ok_result and ok_result._result:
            virsh_vms = ok_result._result.get('stdout_lines', [])
            print(f"List of VMs: {virsh_vms}")

        results_2 = execute_ansible(inventory_path="/etc/fabric/actor/playbooks//inventory",
                                    logger=logger,
                                    playbook_path="/etc/fabric/actor/playbooks//head_vm_provisioning.yml",
                                    extra_vars={"operation": "list", "host": str(host)},
                                    ansible_python_interpreter='/usr/bin/python3.6')

        if results_2 and results_2.get('openstack_servers'):
            servers = results_2.get('openstack_servers')
            for s in servers:
                os_vms[s.get('OS-EXT-SRV-ATTR:instance_name')] = s.get('name')

        # Find extra VMs on Virsh and delete them
        for v in virsh_vms:
            if v not in os_vms:
                results_3 = execute_ansible(inventory_path="/etc/fabric/actor/playbooks//inventory",
                                            logger=logger,
                                            playbook_path="/etc/fabric/actor/playbooks//head_vm_provisioning.yml",
                                            extra_vars={"operation": "delete", "host": str(host)},
                                            ansible_python_interpreter='/usr/bin/python3.6')
                logger.info(f"Deleting for instance: {v} result: {results_3}")


if __name__ == "__main__":
    import time
    tpb = TestPlaybooks()
    #tpb.test_create_vm_success_no_pci()

    #time.sleep(10)
    #tpb.test_delete_vm_success_no_pci()
    #time.sleep(10)
    #u = tpb.test_create_vm_success()

    #time.sleep(10)
    #tpb.test_delete_vm_success(u=u)
    #tpb.test_config_nw_interface_tagged()
    #tpb.test_config_nw_interface()

    #tpb.test_poa_cpuinfo()
    #tpb.test_poa_numainfo()

    #tpb.test_poa_numatune()
    #tpb.test_poa_cpupin()
    #tpb.test_poa_reboot()

    #tpb.test_fpga_prov()

    #test_audit()
    tpb.test_switch()

